﻿//---------------------------------------------------------------------------
//
//   ExportASC
//
//          Copyright(C) 1999-2012, tetraface Inc.
//
//     Sample of Export plug-in for exporting an STL (Stereolithography) 
//     file. Created DLL must be installed in "Plugins\Export" directory.
//
//    　STL(Stereolithography)形式を出力するエクスポートプラグインのサンプル。
//    　作成したDLLは"Plugins\Export"フォルダに入れる必要がある。
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "MQPlugin.h"
#include "MQ3DLib.h"

BOOL SaveSTL(const char *filename, MQDocument doc);


//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	// Nothing to do and return TRUE.
	//プラグインとしては特に必要な処理はないので、何もせずにTRUEを返す
    return TRUE;
}

//---------------------------------------------------------------------------
//  MQGetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT void MQGetPlugInID(DWORD *Product, DWORD *ID)
{
	// Return a product(auther) ID and a plug-in ID.
	// They can be random values that do not overlap with other.
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0x3EA834CB;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQGetPlugInName(void)
{
	// Return a plug-in name
	// プラグイン名
	return "Stereolithography Exporter   Copyright(C) 1999-2012, tetraface Inc.";
}

//---------------------------------------------------------------------------
//  MQGetPlugInType
//    プラグインのタイプを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT int MQGetPlugInType(void)
{
	// This is an Export plug-in.
	// 出力用プラグインである
	return MQPLUGIN_TYPE_EXPORT;
}

//---------------------------------------------------------------------------
//  MQEnumFileType
//    入力または出力可能なファイルタイプを返す。
//    ファイルタイプは別名保存時のダイアログに表示される。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumFileType(int index)
{
	switch(index){
	case 0: return "Stereolithography (*.stl)";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQEnumFileExt
//    入力または出力可能な拡張子を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumFileExt(int index)
{
	switch(index){
	case 0: return "stl";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQExportFile
//    [別名で保存]で保存するときに呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT BOOL MQExportFile(int index, const char *filename, MQDocument doc)
{
	switch(index){
	case 0: return SaveSTL(filename, doc);
	}
	return FALSE;
}


// Export a file using a STL format
// STL形式で出力
BOOL SaveSTL(const char *filename, MQDocument doc)
{
	// Show a dialog for converting axes
	// 座標軸変換用ダイアログの表示
	MQFileDialogInfo dlginfo;
	memset(&dlginfo, 0, sizeof(dlginfo));
	dlginfo.dwSize = sizeof(dlginfo);
	dlginfo.axis_x = MQFILE_TYPE_RIGHT;
	dlginfo.axis_y = MQFILE_TYPE_BACK;
	dlginfo.axis_z = MQFILE_TYPE_UP;
	dlginfo.softname = "";
	MQ_ShowFileDialog("STL Export", &dlginfo);

	// Open a file.
	// ファイルを作成・オープンする
	FILE *fh;
	errno_t ret_s = fopen_s(&fh, filename, "wb");
	if(ret_s != 0){
		return FALSE;
	}

	// Create a temporal object for merging multiple objects.
	// 複数のオブジェクトを１つに合成するために、作業用オブジェクトを作成する
	MQObject obj = MQ_CreateObject();

	// Merge all objects to the temporal object.
	// 作業用オブジェクトにすべてのオブジェクトを合成
	int objcount = doc->GetObjectCount();
	for(int i=0; i<objcount; i++)
	{
		MQObject thisobj = doc->GetObject(i);
		if(thisobj == NULL) 
			continue;
		// Duplicate the object because the contents will be changed by a freezing and so on.
		// フリーズなどにより内容が変更されてしまうので、オブジェクトを複製しておく
		thisobj = thisobj->Clone();
		// Freeze attributes such as a patch.
		// 曲面などの属性をフリーズ
		thisobj->Freeze(MQOBJECT_FREEZE_ALL);
		// Merge to the temporal object.
		// 作業用オブジェクトに合成する
		obj->Merge(thisobj);
		// Delete the duplicated object.
		// 複製したオブジェクトはもう不要なので破棄
		thisobj->DeleteThis();
	}

	// Get a number of vertices and a number of faces.
	// 面と頂点の数を取得
	int face_count = obj->GetFaceCount();
	int vert_count = obj->GetVertexCount();

	// Count a number of output triangles because a STL format cannot have 
	// quadrangles.
	// STL形式は三角形しか扱えないので、三角形面は１つ、四角形面は２つとして、
	// 出力後の三角面の数を数える
	unsigned int triangle_count = 0;
	for(int face_index=0; face_index<face_count; face_index++)
	{
		switch(obj->GetFacePointCount(face_index)){
		case 3: triangle_count++; break;
		case 4: triangle_count+=2; break;
		}
	}

	// Write a header description.
	// ファイルの先頭にデータを書き込み
	char header[80];
	memset(header, 0, sizeof(header));
	sprintf_s(header, sizeof(header), "Created by Metasequoia.");
	fwrite(header, 80, 1, fh);

	// Write a number of triangles.
	// 三角形の数を出力
	fwrite(&triangle_count, 4, 1, fh);

	// Write triangles.
	// 三角形を出力
	for(int face_index=0; face_index<face_count; face_index++)
	{
		int pt_num = obj->GetFacePointCount(face_index);
		if(pt_num < 3) continue;

		// Get indices of vertices.
		// 頂点インデックスの配列を得る
		int index[4];
		obj->GetFacePointArray(face_index, index);

		int indices[2][3];
		int iter;
		switch(pt_num){
		case 3:
			iter = 1;
			indices[0][0] = 0;
			indices[0][1] = 1;
			indices[0][2] = 2;
			break;
		case 4:
			iter = 2;
			{
				// Determine triangles to be divided
				// 三角形に分割する方向を決定
				MQPoint pts[4];
				pts[0] = obj->GetVertex(index[0]);
				pts[1] = obj->GetVertex(index[1]);
				pts[2] = obj->GetVertex(index[2]);
				pts[3] = obj->GetVertex(index[3]);
				MQPoint nrm1 = GetNormal(pts[2], pts[1], pts[0]);
				MQPoint nrm2 = GetNormal(pts[3], pts[2], pts[0]);
				if(GetInnerProduct(nrm1, nrm2) > 0.0f){
					indices[0][0] = 0;
					indices[0][1] = 1;
					indices[0][2] = 2;
					indices[1][0] = 0;
					indices[1][1] = 2;
					indices[1][2] = 3;
				}else{
					indices[0][0] = 0;
					indices[0][1] = 1;
					indices[0][2] = 3;
					indices[1][0] = 1;
					indices[1][1] = 2;
					indices[1][2] = 3;
				}
			}
			break;
		}

		for(int i=0; i<iter; i++)
		{
			float val[3];
			MQPoint pts[3];

			// Get positions of vertices.
			// 頂点の位置を取得
			pts[0] = obj->GetVertex(index[indices[i][0]]);
			pts[1] = obj->GetVertex(index[indices[i][1]]);
			pts[2] = obj->GetVertex(index[indices[i][2]]);

			// Convert an axis coordinate of the vertex.
			// 座標軸をダイアログで指定したものに変換
			MQ_ExportAxis(&dlginfo, pts, 3);

			// Calculate a normal vector.
			// 法線ベクトルを計算
			MQPoint nrm = GetNormal(pts[2], pts[1], pts[0]);

			// Write the normal vector.
			// 法線ベクトルを出力
			val[0] = nrm.x;
			val[1] = nrm.y;
			val[2] = nrm.z;
			fwrite(val, 4, 3, fh);

			// Write the vertices.
			// 頂点座標を出力
			val[0] = pts[0].x;
			val[1] = pts[0].y;
			val[2] = pts[0].z;
			fwrite(val, 4, 3, fh);
			val[0] = pts[1].x;
			val[1] = pts[1].y;
			val[2] = pts[1].z;
			fwrite(val, 4, 3, fh);
			val[0] = pts[2].x;
			val[1] = pts[2].y;
			val[2] = pts[2].z;
			fwrite(val, 4, 3, fh);

			// Write an unused attribute
			static const unsigned char dummy[2] = {0,0};
			fwrite(dummy, 2, 1, fh);
		}
	}

	// Delete the temporal object.
	// 作業用オブジェクトを破棄
	obj->DeleteThis();

	// Close a file.
	// ファイルをクローズ
	int ret = fclose(fh);

	return (ret == 0);
}

